##-----------------------------------------------------------------------
## This function calculates the cost matrix for a homoscedastic model
## G[k,i] is the sum of squared residuals of a segment from i to (i+k-1)
##-----------------------------------------------------------------------
costMatrix = function(x, y, beta, delta, maxk, homThr) {

  if (length(x) != length(y))
	stop("'x' (LRR) and 'y' (BAF) should be of same length.")
  
  n = length(x)
  if(!is.numeric(maxk)||(length(maxk)!=1)||(maxk<1)||(maxk>n))
    stop(sprintf("'maxk' must be a single number between 1 and the number of rows of 'x' and 'y': %d.", n))
  
  ##########  PREPROCESS BAF INTO smBAF ##########
  # mirror BAF in order to get mBAF
  y = abs(y-0.5)+0.5
  # get SNPs in the homozygous band of heterozygous regions
  homosInHeter = intersect(which(y>homThr),which(getHomo(y,germHomozyg.mBAF.thr=homThr)==0))
  # calculate num of non-NA observations up to each SNP
  numHomosInHeter = c()
  numHomosInHeter[1] = 1 * (1%in%homosInHeter)
  for (i in 2:n)
    numHomosInHeter[i] = numHomosInHeter[i-1] + 1 * (i%in%homosInHeter)
  # simplify mBAF to imBAF
  y[homosInHeter] = NA
  
  # smooth slightly the LRR; improves recall rates at small error thresholds (<3SNPs)
  x = filter(x,rep(1/2,2),sides=2,circular=TRUE)
  qx = x*x
  crx = cqx = x
  crx[!is.na(crx)] = cumsum(x[!is.na(x)])
  cqx[!is.na(cqx)] = cumsum(qx[!is.na(qx)])
  qy = y*y
  cry = cqy = y
  cry[!is.na(cry)] = cumsum(y[!is.na(y)])
  cry = approx(which(!is.na(cry)), cry[which(!is.na(cry))], 1:length(cry), method="constant", rule=2)$y 
  cqy[!is.na(cqy)] = cumsum(qy[!is.na(qy)])
  cqy = approx(which(!is.na(cqy)), cqy[which(!is.na(cqy))], 1:length(cqy), method="constant", rule=2)$y 
  
  G = G2 = D = matrix(as.numeric(NA), nrow=maxk, ncol=n)
  k = 1:maxk
  G[, 1] = cqx[k] - crx[k]*crx[k]/k
  G2[, 1] = cqy[k] - cry[k]*cry[k]/k
  for(k in 1:(maxk-1)) {
    i   = 1:(n-k)
    j   = 2:(n-k+1)
    cqkx = cqx[i+k]-cqx[i]
    crkx = crx[i+k]-crx[i]
	cqky = cqy[i+k]-cqy[i]
    crky = cry[i+k]-cry[i]
    G[k,j] = cqkx - crkx*crkx/k
	# use num of observations instead of length, which would include NAs
	len = k - (numHomosInHeter[i+k] - numHomosInHeter[i])
    G2[k,j] = cqky - crky*crky/len
	# get SS of regions w/o info through interpolation
	G2[k,which(G2[k,]==Inf)] = NA
	G2[k,] = approx(which(!is.na(G2[k,])), G2[k,which(!is.na(G2[k,]))], 1:n, method="linear", rule=2)$y 
	# length normalization
	# G2[k,j] = G2[k,j] * (len/k)
	G[k,j] = G[k,j] * (k/(k+len))
	G2[k,j] = G2[k,j] * (len/(k+len))
  }
  G = G - min(G,na.rm=TRUE)
  G2 = G2 - min(G2,na.rm=TRUE)
	
  D = (G**delta+beta*G2**delta)^(1/delta)
  return(D)
}

##################################################################
####
#### method by Popova et al., as included in the GAP software
####
##################################################################

getHomo = function(data2, germHomozyg.mBAF.thr=0.95, germHomozyg.length.min=50, germHomozyg.outlier=1/15) {

##################################################################
####
#### Processing BAF in order to get mBAF, germline Homosygous SNPs and germline LOH regions
####
##################################################################

nSNP = length(data2)

#### Transforming BAF to major BAF

dataF<-matrix(0, ncol=3, nrow=nSNP)
C_mBAF = 1
C_gHomo = 2
C_gLOH = 3

dataF[,C_mBAF]<-data2

#### Set NAs in BAF to 1

tt<-which(is.na(dataF[,C_mBAF]))
if(length(tt)>0){dataF[tt,C_mBAF]<-1}

############################################
####  Determining germline homozygous regions:
#### regions with preliminary homozygous SNPs of >=50 length
####      (<=2 outliers for 30 SNP window are allowed:
####             here germHomozyg.outlier=(1/15))
#### Heterozygous regions are determined by at least 3 Hetero SNPs in 20 SNP window
############################################

########################################
#### 1. Marking germline homozygous SNPs 

dataF[,C_gHomo]<-0
tt<-which(dataF[,C_mBAF]>germHomozyg.mBAF.thr)
if(length(tt)>0){dataF[tt,C_gHomo]<-1}

########################################
#### 2.  Heterozygous SNPs surrounded by at least 10 homozygous are set to Homozygous

dataF[,C_gLOH]<-1-dataF[,C_gHomo]

for(k in 1:10){
	dataF[(1+k):nSNP,C_gLOH]<-dataF[(1+k):nSNP,C_gLOH]+(1-dataF[1:(nSNP-k),C_gHomo])
	dataF[1:(nSNP-k),C_gLOH]<-dataF[1:(nSNP-k),C_gLOH]+(1-dataF[(1+k):nSNP,C_gHomo])
}

tt<-as.numeric(which(dataF[,C_gLOH]*(1-dataF[,C_gHomo])==1))

dataF[tt,C_gHomo]<-1


########################################
#### 3.  2 Heterozygous SNPs in 31 homozygous window are set to Homozygous

dataF[,C_gLOH]<-1-dataF[,C_gHomo]

for(k in 1:15){
	dataF[(1+k):nSNP,C_gLOH]<-dataF[(1+k):nSNP,C_gLOH]+(1-dataF[1:(nSNP-k),C_gHomo])
	dataF[1:(nSNP-k),C_gLOH]<-dataF[1:(nSNP-k),C_gLOH]+(1-dataF[(1+k):nSNP,C_gHomo])
}

tt<-as.numeric(which(dataF[,C_gLOH]*(1-dataF[,C_gHomo])==2))

dataF[tt,C_gHomo]<-1

#######################################
#### 4. Determining homozygous regions

dataF[,C_gLOH]<-dataF[,C_gHomo]

for(k in 1:2){
	dataF[(1+k):nSNP,C_gLOH]<-dataF[(1+k):nSNP,C_gLOH]*dataF[1:(nSNP-k),C_gHomo]
	dataF[1:(nSNP-k),C_gLOH]<-dataF[1:(nSNP-k),C_gLOH]*dataF[(1+k):nSNP,C_gHomo]
}

##### tmp matrix of homozygous regions

tmp<-c(2:nSNP)*sign(abs(dataF[2:nSNP,C_gLOH]-dataF[1:(nSNP-1),C_gLOH]))
# +
# abs(dataF[2:nSNP,]-dataF[1:(nSNP-1),2])+abs(dataF[2:nSNP,3]-dataF[1:(nSNP-1),3]))


if(length(which(tmp==0))!=0){tmp<-tmp[-which(tmp==0)]}
tmp<-c(1,tmp)
tmp<-as.numeric(rownames(table(tmp)))

tmp<-cbind(tmp,matrix(0,length(tmp),5))

colnames(tmp)<-c("Ind","Ind_K","Len","Chr","nGH_in","nGH_up")

tmp[1:(dim(tmp)[[1]]-1),2]<-tmp[2:dim(tmp)[[1]],1]-1
tmp[dim(tmp)[[1]],2]<-nSNP
# tmp[,4]<-dataF[tmp[,2],C_Chr]+(dataF[tmp[,2],C_Arm]-1)/2
tmp[,5]<-dataF[tmp[,2],C_gLOH]

tt<-which(tmp[,5]==0);if(length(tt)!=0){tmp<-tmp[-tt,,drop=FALSE]}
tt<-which(tmp[,1]>2);if(length(tt)!=0){tmp[tt,1]<-tmp[tt,1]-2}
tt<-which(tmp[,2]<(nSNP-2));if(length(tt)!=0){tmp[tt,2]<-tmp[tt,2]+2}
tmp[,3]<-tmp[,2]-tmp[,1]+1


FL<-T
while(FL==T){
	FL<-F
	k<-2
	for(k in 1:(dim(tmp)[[1]]-1)){
		tmp[k,5]<-sum(-dataF[tmp[k,1]:tmp[k,2],C_gHomo]+1)
		tmp[k,6]<-sum(-dataF[(tmp[k,2]+1):(tmp[k+1,1]-1),C_gHomo]+1)

		# if (!is.na(tmp[k+1,4]))
		# if(tmp[k,4]==tmp[k+1,4]){
		if((tmp[k,6]<4)&&((tmp[k+1,5]+tmp[k,6])/(tmp[k+1,2]-tmp[k,2]+1)<germHomozyg.outlier)&&
              ((tmp[k,5]+tmp[k+1,5]+tmp[k,6])/(tmp[k+1,2]-tmp[k,1]+1)<germHomozyg.outlier)){
			tmp[k+1,1]<-tmp[k,1]
			tmp[k+1,3]<-tmp[k+1,2]-tmp[k+1,1]+1
			tmp[k,1]<-0
			FL<-T
		}
		# }
	}#k

	if(length(which(tmp[,1]==0))!=0){
	tmp<-tmp[-which(tmp[,1]==0),]}

}#FL

tt<-which(tmp[,3]<germHomozyg.length.min);if(length(tt)!=0){tmp<-tmp[-tt,,drop=FALSE]}

#######################################
#### 5. Saving filtered homozygous regions in dataF C_gLOH column

dataF[,C_gLOH]<-0

if(dim(tmp)[[1]]>0)
for(k in 1:dim(tmp)[[1]]){
	dataF[tmp[k,1]:tmp[k,2],C_gLOH]<-1
}#k

rm(tmp)

dataF[,3]
}
